#include "JuliaFractalPixelShader.h"
#include <d3d9.h>
#include <d3dx9.h>

//Engine objects and states
static IDirect3D9* D3DInterface=NULL;
static IDirect3DVertexBuffer9* VertextBuffer=NULL;
static IDirect3DDevice9* D3DDevice=NULL;
static int CurrentMaxIterations; //The current number of max iterations in the pixel shader

const char* PrepareRenderEngine(HWND hWnd, DWORD Width, DWORD Height, bool IsImmediate) //Create the engine objects
{
	ReleaseRenderEngine(); //If engine objects have already been created, free them before recreating

	//Create the DX interface
	D3DInterface=Direct3DCreate9(D3D_SDK_VERSION);
	if(!D3DInterface)
		return "Cannot create DirectX Interface";

	//Determine if we have the needed capabilities, and if not, set to render via software
	D3DDEVTYPE dev_type=D3DDEVTYPE_HAL;
	D3DCAPS9 d3dcaps9;
	ZeroMemory(&d3dcaps9, sizeof(d3dcaps9));
	if(FAILED(D3DInterface->GetDeviceCaps(D3DADAPTER_DEFAULT, dev_type, &d3dcaps9)) || d3dcaps9.PixelShaderVersion<D3DPS_VERSION(2,1))
		dev_type=D3DDEVTYPE_REF;

	//Get the current display mode
	D3DDISPLAYMODE displayMode;
	ZeroMemory(&displayMode, sizeof(displayMode));
	if(FAILED(D3DInterface->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &displayMode)))
		return "Cannot query display mode";

	//Create the DX Device
	D3DPRESENT_PARAMETERS d3dpp;
	ZeroMemory(&d3dpp, sizeof(d3dpp));
	d3dpp.BackBufferFormat=displayMode.Format;
	d3dpp.Windowed=true;
	d3dpp.SwapEffect=D3DSWAPEFFECT_DISCARD;
	d3dpp.EnableAutoDepthStencil=FALSE;
	d3dpp.PresentationInterval=(IsImmediate ? D3DPRESENT_INTERVAL_IMMEDIATE : D3DPRESENT_INTERVAL_DEFAULT);
	if(FAILED(D3DInterface->CreateDevice(D3DADAPTER_DEFAULT, dev_type, hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &D3DDevice)))
		return "Cannot create DirectX Device";

	//Define a full screen quad
	const int VERTEX_FORMAT=(D3DFVF_XYZ|D3DFVF_TEX1);
	float HWidth=(float)Width/2, HHeight=(float)Height/2;
	Vertex Vertices[4] =
	{
		-HWidth, HHeight,1,0,0,
		 HWidth, HHeight,1,1,0,
		 HWidth,-HHeight,1,1,1,
		-HWidth,-HHeight,1,0,1
	};

	//Create the vertex buffer
	void* ptr;
	if(
		FAILED(D3DDevice->CreateVertexBuffer(4*sizeof(Vertex), D3DUSAGE_WRITEONLY, VERTEX_FORMAT, D3DPOOL_DEFAULT, &VertextBuffer, NULL)) ||
		FAILED(VertextBuffer->Lock(0, 4*sizeof(Vertex), &ptr, 0))
	)
		return "Cannot create vertex buffer";
	memcpy(ptr, Vertices, 4*sizeof(Vertex));
	if(
		FAILED(VertextBuffer->Unlock()) ||
		FAILED(D3DDevice->SetFVF(VERTEX_FORMAT)) ||
		FAILED(D3DDevice->SetStreamSource(0, VertextBuffer, 0, sizeof(Vertex)))
	)
		return "Cannot populate vertex buffer";

	//Create and set the pixel shader
	LPD3DXBUFFER pCode, pErrorMsgs=NULL;
	IDirect3DPixelShader9* PixelShader;
	if(FAILED(D3DXAssembleShaderFromResource(NULL, "Pixel_Shader", NULL, NULL, 0, &pCode, &pErrorMsgs)))
	{
		if(!pErrorMsgs)
			return "Cannot find pixel shader file";
		MessageBox(NULL, (char*)pErrorMsgs->GetBufferPointer(), "Pixel Shader Error", MB_OK);
		return "Cannot create pixel shader";
	}
	if(
		FAILED(D3DDevice->CreatePixelShader((DWORD*)pCode->GetBufferPointer(), &PixelShader)) ||
		FAILED(D3DDevice->SetPixelShader(PixelShader))
	)
		return "Cannot initate pixel shader";
	PixelShader->Release();

	//Render states
	D3DDevice->SetRenderState(D3DRS_LIGHTING, FALSE); //No lighting

	//Prepare a 2D screen
	D3DXMATRIX Ortho2D;
	D3DXMATRIX Identity;
	D3DXMatrixOrthoLH(&Ortho2D, (float)Width, (float)Height, 0.0f, 1.0f);
	D3DXMatrixIdentity(&Identity);
	D3DDevice->SetTransform(D3DTS_PROJECTION, &Ortho2D);
	D3DDevice->SetTransform(D3DTS_WORLD, &Identity);
	D3DDevice->SetTransform(D3DTS_VIEW, &Identity);

	//Create the color range palette as a bitmap
	const DWORD ColorRanges[]={
		0x000080, //Dark Blue
		0x0000FF, //Blue
		0xA0A0FF, //Light Blue
		0xFFFF80, //Light yellow
		0x80FF80, //Light green
		0x008000, //Dark green
		0x008080, //Forest green
		0x00FFFF, //Cyan
		0xFF00FF, //Purple
		0xFF0000, //Red
		0x800000, //Dark red
		0x000080 //Dark blue - Start back at the beginning
	}, NumColorRanges=sizeof(ColorRanges)/sizeof(DWORD);
	const int FHS=sizeof(BITMAPFILEHEADER),IHS=sizeof(BITMAPINFOHEADER); //File header and info header sizes
	BITMAPFILEHEADER BFH={0x4D42, FHS+IHS+sizeof(ColorRanges), 0, 0, FHS+IHS};
	BITMAPINFOHEADER BIH={IHS, NumColorRanges, 1, 1, 32, BI_RGB,	sizeof(ColorRanges), 0, 0, 0, 0};
	BYTE *MyPixelBitmap=new BYTE[FHS+IHS+sizeof(ColorRanges)];
	memcpy(MyPixelBitmap, &BFH, FHS);
	memcpy(MyPixelBitmap+FHS, &BIH, IHS);
	memcpy(MyPixelBitmap+FHS+IHS, ColorRanges, sizeof(ColorRanges));

	//Copy the bitmap to a texture
	IDirect3DTexture9* PaletteTexture;
	if(
		FAILED(D3DXCreateTextureFromFileInMemoryEx(D3DDevice, MyPixelBitmap, FHS+IHS+sizeof(ColorRanges), NumColorRanges, 1, 1, 0, D3DFMT_X8R8G8B8, D3DPOOL_DEFAULT, D3DX_FILTER_LINEAR, D3DX_FILTER_LINEAR, 0, NULL, NULL, &PaletteTexture)) ||
		FAILED(D3DDevice->SetTexture(0, PaletteTexture)) ||
		FAILED(D3DDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR)) ||
		FAILED(D3DDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR))
	)
		return "Cannot create palette";
	PaletteTexture->Release();
	delete[] MyPixelBitmap;

	//Set the pixel shader constants
	CurrentMaxIterations=-555; //Set as currently unknown
	const int RangeSize=20;
	const float PSFConstants[]={
		-555, -555, -555, -555, //XWidth=XMax-XMin, YHeight=YMax-YMin, YMin, XMin (All Currently unknown)
		1,0,1,-1, //Dot product multiply for xtemp
		0,1,0,2, //Dot product multiply for y
		1,1,0,(float)CurrentMaxIterations, //xyz=Dot product for Complex#^2 breakout test, w=MaxIterations (EscapeLevel)
		4,1.0f/(RangeSize*NumColorRanges),0.32663425997828f,-1.4426950408889f, //BailoutRadius^2, ClampToTexture=1/(RangeSize*NumColorRanges), log(2*log(BailoutRadius)) [used for escape lerp], -1/log(2) [used for escape lerp]
		0.6931471806f,-555,-555,-555 //ReverseLogBase2=ln(2),JuliaVarsUseC,JuliaVarsReal,JuliaVarsImag
	};
	D3DDevice->SetPixelShaderConstantF(0, PSFConstants, 6);

	return NULL;
}

void ReleaseRenderEngine() //Release the engine objects
{
	if(VertextBuffer)
		VertextBuffer->Release();
	if(D3DDevice)
		D3DDevice->Release();
	if(D3DInterface)
		D3DInterface->Release();
	VertextBuffer=NULL;
	D3DDevice=NULL;
	D3DInterface=NULL;
}

void DrawFractal(Fractal F, bool DrawToScreen) //Runs a render pass
{
	//Send fractal render space to the pixel shader
	float PSFConstants[]={(float)(F.XMax-F.XMin), (float)(F.YMax-F.YMin), (float)F.YMin, (float)F.XMin}; //XWidth=XMax-XMin, YHeight=YMax-YMin, YMin, XMin
	D3DDevice->SetPixelShaderConstantF(0, PSFConstants, 1);

	//Send fractal config to the pixel shader
	float PSFConstants2[]={0.6931471806f,F.UseC ? 1.0f : 0, (float)F.AddReal, (float)F.AddImag}; //ReverseLogBase2=ln(2),JuliaVarsUseC,JuliaVarsReal,JuliaVarsImag
	D3DDevice->SetPixelShaderConstantF(5, PSFConstants2, 1);

	//If MaxIterations has changed then update it
	if(CurrentMaxIterations!=F.MaxIterations)
	{
		const float PSFConstants[]={1,1,0,(float)F.MaxIterations};
		const int PSIConstants[]={F.MaxIterations,0,0,0}; //MaxIterations,0,0,0
		D3DDevice->SetPixelShaderConstantF(3, PSFConstants, 1);
		D3DDevice->SetPixelShaderConstantI(0, PSIConstants, 1);
		CurrentMaxIterations=F.MaxIterations;
	}

	//Draw the fractal
	D3DDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_COLORVALUE(0.5, 0.5, 0.5, 1.0), 1.0, 0);
	D3DDevice->BeginScene();
	D3DDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, 0, 2);
	D3DDevice->EndScene();
	if(DrawToScreen)
		D3DDevice->Present(NULL, NULL, NULL, NULL);
}

void* GetD3DDevice() { return D3DDevice; } //Returns the IDirect3DDevice9*